# -*- coding: utf-8 -*-
from shapely.geometry import Polygon
from geographiclib.geodesic import Geodesic
import pandas as pd
import numpy as np


_LARGE_BOX_WIDTH = 2000


def _get_box(latlon0, azimuth, length, width=_LARGE_BOX_WIDTH, offset=0):
    """Create a single box."""

    start = direct_geodetic(latlon0, azimuth, offset)
    azis = ((azimuth - 90) % 360, azimuth,
            (azimuth + 90) % 360, (azimuth + 180) % 360)
    dists = (width/2, length, width, length)
    latlon = start
    corners = []
    for a, d in zip(azis, dists):
        latlon = direct_geodetic(latlon, a, d)
        corners.append(latlon[::-1])
    box = {'poly': Polygon(corners),
           'length': length,
           'pos': offset + length/2,
           'latlon': direct_geodetic(start, azimuth, length/2)}
    return box


def _get_box_cartesian(latlon0, azimuth, length, width=2000, offset=0):
    """Create a single box."""

    import math
    d = offset
    a = azimuth
    lat = d*math.cos(a*math.pi/180) / 111.195
    lon = d*math.sin(a*math.pi/180) / 111.195
    start = (latlon0[0]+lat, latlon0[1]+lon)
    azis = ((azimuth - 90) % 360, azimuth,
            (azimuth + 90) % 360, (azimuth + 180) % 360, (azimuth + 270) % 360)
    dists = (width/2, length/2, width, length, width)
    latlon = start
    corners = []
    for a, d in zip(azis, dists):
        # latlon = direct_geodetic(latlon, a, d)
        # in km
        lat = d*math.cos(a*math.pi/180) / 111.195
        lon = d*math.sin(a*math.pi/180) / 111.195
        latlon = (latlon[0]+lat, latlon[1]+lon)
        corners.append(latlon[::-1])

    lat = length/2*math.cos(azimuth*math.pi/180) / 111.195
    lon = length/2*math.sin(azimuth*math.pi/180) / 111.195
    box = {'poly': Polygon(corners),
           'length': length,
           'pos': offset + length/2,
           'latlon': (start[0], start[1]+lon)}
#            'latlon': direct_geodetic(start, azimuth, length/2)}
    return box


def direct_geodetic(latlon, azi, dist):
    """
    Solve direct geodetic problem with geographiclib.

    :param tuple latlon: coordinates of first point
    :param azi: azimuth of direction
    :param dist: distance in km

    :return: coordinates (lat, lon) of second point on a WGS84 globe
    """
    coords = Geodesic.WGS84.Direct(latlon[0], latlon[1], azi, dist * 1000)
    return coords['lat2'], coords['lon2']


def get_profile_boxes(latlon0, azimuth, bins, width=_LARGE_BOX_WIDTH):
    """
    Create 2D boxes for usage in `profile()` function.

    :param tuple latlon0: coordinates of starting point of profile
    :param azimuth: azimuth of profile direction
    :param tuple bins: Edges of the distance bins in km (e.g. (0, 10, 20, 30))
    :param width: width of the boxes in km (default: large)
    :return: List of box dicts. Each box has the entries
        'poly' (shapely polygon with lonlat corners),
        'length' (length in km),
        'pos' (midpoint of box in km from starting coordinates),
        'latlon' (midpoint of box as coordinates)
    """
    boxes = []
    for i in range(len(bins)-1):
        length = bins[i+1] - bins[i]
        # box = _get_box(latlon0, azimuth, length, width, offset=bins[i])
        box = _get_box_cartesian(latlon0, azimuth, length, width, offset=bins[i])
        if i == 0:
            box['profile'] = {}
            box['profile']['latlon'] = latlon0
            box['profile']['azimuth'] = azimuth
            box['profile']['length'] = bins[-1] - bins[0]
            box['profile']['width'] = width
        boxes.append(box)
    return boxes


def _find_box(latlon, boxes, crs=None):
    """Return the box which encloses the coordinates."""
    import cartopy.crs as ccrs
    from shapely.geometry import Point
    if crs is None:
        latlons = [boxes[len(boxes)//2]['latlon']]
        latlon0 = np.median(latlons, axis=0)
        crs = ccrs.AzimuthalEquidistant(*latlon0[::-1])
    pc = ccrs.PlateCarree()
    p = crs.project_geometry(Point(*latlon[::-1]), pc)
    for box in boxes:
        poly = crs.project_geometry(box['poly'], pc)
        if p.within(poly):
            return box


def _find_box_cartesian(latlon, boxes):
    """Return the box which encloses the coordinates."""
    from shapely.geometry import Point
    p = Point(latlon[::-1])
    for box in boxes:
        poly = box['poly']
        if p.within(poly):
            return box


def profile(stream, boxes, crs=None):
    """
    Stack traces in stream by piercing point coordinates in defined boxes.

    :param stream: stream with pre-calculated piercing point coordinates
    :param boxes: boxes created with `get_profile_boxes()`
    :param crs: cartopy projection (default: AzimuthalEquidistant)
    :return: profile stream
    """

    # read all migrated data saved in trace headers into Pandas DataFrame
    dfmig = pd.DataFrame()
    for tr in stream:
        df = tr.stats.mig
        dfmig = dfmig.append(df)

    # get depth samplings
    depths = dfmig["depth"].unique()
    depths = depths.tolist()

    # get lateral samplings
    pos_dist = []
    for b in boxes:
        pos_dist.append(b["pos"])
    # print(pos_dist)

    # sampling number
    npos = len(pos_dist)
    ndep = len(depths)
    # init numpy array, stack - final stacked CRP image, count - hit number in each cell
    stack = np.zeros((ndep, npos), dtype=float)
    count = np.zeros((ndep, npos), dtype=int)

    # loops
    for depth in depths:
        df2 = dfmig[dfmig["depth"] == depth]

        lats = df2["lat"].to_numpy()
        lons = df2["lon"].to_numpy()
        data = df2["data"].to_numpy()

        for i, lat in enumerate(lats):
            lon = lons[i]
            ppoint = (lat, lon)
            box = _find_box_cartesian(ppoint, boxes)
            if box is None:
                continue
            pos = box['pos']

            idep = depths.index(depth)
            ipos = pos_dist.index(pos)
            # print(pos, ipos, ppoint)

            stack[idep][ipos] += data[i]
            count[idep][ipos] += 1

    d = np.array(depths)
    p = np.array(pos_dist)

    return p, d, stack



# -*- coding: utf-8 -*-
# Copyright 2019- Weijia Sun, MIT license
"""
Utility functions and classes for Auto and Cross Correlogram calculation.
"""
import scipy.signal
import numpy as np
import collections
from obspy import UTCDateTime

def smooth(x, window_len=None, window='flat', method='zeros'):
    """Smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.

    :param x: the input signal (numpy array)
    :param window_len: the dimension of the smoothing window; should be an
        odd integer
    :param window: the type of window from 'flat', 'hanning', 'hamming',
        'bartlett', 'blackman'
        flat window will produce a moving average smoothing.
    :param method: handling of border effects\n
        'zeros': zero padding on both ends (len(smooth(x)) = len(x))\n
        'reflect': pad reflected signal on both ends (same)\n
        'clip': pad signal on both ends with the last valid value (same)\n
        None: no handling of border effects
        (len(smooth(x)) = len(x) - len(window_len) + 1)
    """
    if window_len is None:
        return x
    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")
    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")
    if window_len < 3:
        return x
    if window not in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is one of 'flat', 'hanning', 'hamming',"
                         "'bartlett', 'blackman'")
    if method == 'zeros':
        s = np.r_[np.zeros((window_len - 1) // 2), x,
                  np.zeros(window_len // 2)]
    elif method == 'reflect':
        s = np.r_[x[(window_len - 1) // 2:0:-1], x,
                  x[-1:-(window_len + 1) // 2:-1]]
    elif method == 'clip':
        s = np.r_[x[0] * np.ones((window_len - 1) // 2), x,
                  x[-1] * np.ones(window_len // 2)]
    else:
        s = x

    if window == 'flat':
        w = np.ones(window_len, 'd')
    else:
        w = getattr(np, window)(window_len)

    return scipy.signal.fftconvolve(w / w.sum(), s, mode='valid')


class IterMultipleComponents(object):
    """
    Return iterable to iterate over associated components of a stream.

    :param stream: Stream with different, possibly many traces. It is
        split into substreams with the same seed id (only last character
        i.e. component may vary)
    :type key: str or None
    :param key: Additionally, the stream is grouped by the values of
         the given stats entry to differentiate between e.g. different events
         (for example key='starttime', key='onset')
    :type number_components: int, tuple of ints or None
    :param number_components: Only iterate through substreams with
         matching number of components.
    """

    def __init__(self, stream, key=None, number_components=None):
        substreams = collections.defaultdict(stream.__class__)
        for tr in stream:
            k = (tr.id[:-1], str(tr.stats[key]) if key is not None else None)
            substreams[k].append(tr)
        n = number_components
        self.substreams = [s for _, s in sorted(substreams.items())
                           if n is None or len(s) == n or
                           (not isinstance(n, int) and len(s) in n)]

    def __len__(self):
        return len(self.substreams)

    def __iter__(self):
        for s in self.substreams:
            yield s


def iter_time(tr, length=3600, overlap=1800):

    tr2 = tr.copy()
    starttime = int(tr2.stats.starttime.timestamp / length) * length
    endtime = int(tr2.stats.endtime.timestamp / length) * length

    time_series = []
    for t in range(starttime, endtime, overlap):
        t1 = UTCDateTime(t)
        t2 = UTCDateTime(t1 + length)
        time_series.append([t1, t2])

    return time_series




# -*- coding: utf-8 -*-
import os
import glob
import commentjson
from acc.io import _load_json, _get_event_id_tr, _get_station_id, _get_noise_id
from acc.core import spectral_whitening, time_norm, rotation, remove_response
from acc.util import iter_time, IterMultipleComponents
from acc.core import _find_start_end_time

from obspy.signal.filter import envelope
from obspy import read, Stream, Trace

from scipy.signal import correlate
import numpy as np

from tqdm import tqdm
from functools import partial
from multiprocessing import Pool
import multiprocessing
import logging

logging.basicConfig(level=logging.NOTSET, format='%(asctime)s %(filename)s[line:%(lineno)d] %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                    filename="acc.processing.log", filemode="a"
                    )


def processing_event_Z(jsonfile):
    """
    processing vertical component of event data

    :param jsonfile:
    :return:
    """

    # load imported data from files
    kwargs = _load_json(jsonfile)
    njobs = kwargs["njobs"]
    datapath = kwargs["io"]["outpath"] + "/0_raw"

    logging.info("This may take a lot of time if dataset is huge.")

    # get filenames in Unix style for further processing
    files = glob.glob(datapath + "/*/*.pkl")
    logging.info("%d files found.", len(files))

    # processing includes downsampling, detrend, demean,
    # instrument response removal, spectral whitening, temporal normalization
    # autocorrelation and filter, then output results.
    do_work = partial(_proc_event_Z, **kwargs)

    numbers = []
    if njobs == 1:
        logging.info('do work sequential (%d cores)', njobs)
        for file in tqdm(files, total=len(files)):
            num = do_work(file)
            numbers.append(num)
    else:
        logging.info('do work parallel (%d cores)', njobs)
        pool = multiprocessing.Pool(njobs)
        for num in tqdm(pool.imap_unordered(do_work, files), total=len(files)):
            numbers.append(num)
        pool.close()
        pool.join()

    logging.info("%d/%d files processed.", sum(numbers), len(files))


def processing_event_full(jsonfile):
    # load imported data from files
    kwargs = _load_json(jsonfile)
    njobs = kwargs["njobs"]
    datapath = kwargs["io"]["outpath"] + "/0_raw"

    logging.info("This may take a lot of time and consume internal memory if dataset is huge.")

    # get filenames in Unix style for further processing
    files = glob.glob(datapath + "/*/*.pkl")
    logging.info("%d files found.", len(files))

    # read files
    st = Stream()
    for file in files:
        st += read(file)
    sampling_rate = kwargs["preprocessing"]["sampling_rate"]
    # processing includes downsampling, detrend, demean, a large dataset
    st = _simple_proc(st, sampling_rate=sampling_rate, njobs=njobs)
    # instrument response removal, spectral whitening, temporal normalization
    # autocorrelation and filter, then output results.
    _proc_event_full(st, **kwargs)


def _proc_event_full(st, **kwargs):
    """
    processings including

    :param st:
    :param kwargs:
    :return:
    """
    # instrument response removal, spectral whitening, temporal normalization
    # autocorrelation and filter, then output results.
    def iter3c(stream):
        # for an event, there is always "onset"
        return IterMultipleComponents(stream, key="onset", number_components=(2, 3))

    # resp removal, rotation, spectral whitening, temporal normalization
    tasks = iter3c(st)
    # loop over streams in each stream containing the 3-component traces
    do_work = partial(_proc_event_rst, **kwargs)

    njobs = kwargs["njobs"]

    numbers = []
    logging.info("deep processing for full event correlogram.")
    print("deep processing for full event correlogram.")
    if njobs == 1:
        logging.info('do work sequential (%d cores)', njobs)
        for task in tqdm(tasks, total=len(tasks)):
            num = do_work(task)
            numbers.append(num)
    else:
        logging.info('do work parallel (%d cores)', njobs)
        pool = multiprocessing.Pool(njobs)
        for num in tqdm(pool.imap_unordered(do_work, tasks), total=len(tasks)):
            numbers.append(num)
        pool.close()
        pool.join()

    logging.info("%d/%d files processed.", sum(numbers), len(tasks))


def _proc_event_rst(st, **kwargs):
    """
    processing including rotation, spectral whitening and temporal normalization.
    and cross-correlation.

    :param st:
    :param kwargs:
    :return: 1 if processing successfully.

    .. Note::
        rst is short as Rotation Spectral whitening and Temporal normalization.
    """
    # calculation SNR
    # trace = Trace()
    for tr in st:
        if tr.stats.channel[-1] in ["Z", "L"]:
            trace = tr.copy()
    snr = _cal_event_snr(tr=trace)
    snr_threshold = kwargs["data_selection_event"]["snr_threshold"]
    if snr < snr_threshold:
        return 0
    for tr in st:
        tr.stats.update({"snr": snr})

    # remove instrumental response
    if kwargs["rm_resp_on"]:
        option_resp = kwargs["preprocessing"]["resp"]
        st2 = Stream()
        for tr in st:
            tr = remove_response(trace=tr, **option_resp)
            st2.append(tr)
        st = st2.copy()

    # st: stream containing 3-comp traces of an event recorded by a station
    # Rotation, Spectral whitening, and Temporal normalization
    rotation_method = kwargs["preprocessing"]["rotation_method"]
    st = _rot(st, method=rotation_method)

    # spectral whitening
    if kwargs["whiten_on"]:
        option_whiten = kwargs["preprocessing"]["whiten"]
        st2 = Stream()
        for tr in st:
            tr = spectral_whitening(tr=tr, **option_whiten)
            st2.append(tr)
        st = st2.copy()

    # temporal normalization
    if kwargs["time_norm_on"]:
        option_time_norm = kwargs["preprocessing"]["time_norm"]
        st2 = Stream()
        for tr in st:
            tr = time_norm(tr, **option_time_norm)
            st2.append(tr)
        st = st2.copy()

    # correlation
    # trace_l = Trace()
    # trace_q = Trace()
    # trace_t = Trace()
    for tr in st:
        if tr.stats.channel[-1] in ["Z", "L"]:
            trace_l = tr.copy()
        elif tr.stats.channel[-1] in ["R", "Q"]:
            trace_q = tr.copy()
        elif tr.stats.channel[-1] in ["T"]:
            trace_t = tr.copy()
        else:
            logging.warn("unknown trace channel.", tr.stats.channel)

    # auto and cross correlation
    # auto
    options = kwargs["correlate"]
    cc_ll = _crosscorrelation(tr1=trace_l, tr2=trace_l, **options)
    # cross
    options = kwargs["cross_correlate"]
    cc_ql = _crosscorrelation(tr1=trace_q, tr2=trace_l, **options)
    cc_tl = _crosscorrelation(tr1=trace_t, tr2=trace_l, **options)

    trace_l = cc_ll.copy()
    trace_q = cc_ql.copy()
    trace_t = cc_tl.copy()

    # rename channel name
    if rotation_method == "NE->RT":
        chn = ["ZZ", "RZ", "TZ"]
    elif rotation_method == "ZNE->LQT":
        chn = ["LL", "QL", 'TL']
    trace_l.stats.channel = chn[0]
    trace_q.stats.channel = chn[1]
    trace_t.stats.channel = chn[2]

    # write into disk
    for tr in [trace_l, trace_q, trace_t]:
        trace_id = tr.id
        station_id = _get_station_id(tr)
        event_id = _get_event_id_tr(tr)
        filename = trace_id + "_" + event_id + ".pkl"
        outpath = kwargs["io"]["outpath"] + "/1_results"
        filepath = "/".join([outpath, station_id])
        # if not os.path.exists(filepath):
        try:
            os.makedirs(filepath)
        except:
            pass

        # extracting data, to have the same length of output data
        tlen = kwargs["correlate"]["window"]
        tlen = abs(tlen[1] - tlen[0])
        tr.trim(starttime=tr.stats.starttime, endtime=tr.stats.starttime + tlen, fill_value=0, pad=True)

        filen = filepath + "/" + filename
        force = kwargs["io"]["force"]
        if not force and os.path.exists(filen):
            pass
        else:
            tr.write(filen, format="PICKLE")
    return 1


def _rot(st, method="NE->RT"):
    """
    rotation.

    :param st:
    :param method: "NE->RT", "ZNE->LQT".
    :return:

    .. Note::
        the code can handle 'Z12', '123', 'ZNE'. It requires trace header of component_azimuth and component_inclination.
        if method == "NE->RT", then the two components rotation will be applied.
        But while the components are "123" or "Z12", they are rotated to "ZNE" first.
    """

    st3c = st.copy()

    # the rotating traces should have the same starttime and endtime
    tmin, tmax = _find_start_end_time(stream=st3c)
    if tmin > tmax:
        return 0
    st3c.trim(starttime=tmin, endtime=tmax)

    # handling the borehole components 1, 2, Z or 1, 2, 3.
    comp = []
    for tr in st3c:
        comp.append(tr.stats.channel[-1])
    comp.sort()
    cc = "".join(comp)
    if cc == "12Z":
        cc = "Z12"
    # rotate2zne
    if cc in ["123", "Z12"]:
        zne = rotate2zne(st3c[0].data, st3c[0].stats.component_azimuth, st3c[0].stats.component_inclination,
                         st3c[1].data, st3c[1].stats.component_azimuth, st3c[1].stats.component_inclination,
                         st3c[2].data, st3c[2].stats.component_azimuth, st3c[2].stats.component_inclination
                         )

        for tr, new_data, component in zip(st3c, zne, "ZNE"):
            tr.data = new_data
            tr.stats.channel = tr.stats.channel[:-1] + component

    # rotate to various coordinates
    # import matplotlib.pyplot as plt
    # print(st3c)
    # tr = st3c.select(channel="BHZ")[0]
    # fig, ax = plt.subplots(2, 1, sharex=True)
    # ax[0].plot(tr.times(), tr.data)
    st3c.rotate(method=method)
    # tr = st3c.select(channel="BHL")[0]
    # ax[0].plot(tr.times(), tr.data)
    # ax[0].set_xlim([670, 700])
    # plt.tight_layout()
    # plt.show()
    # os._exit(0)
    return st3c


def _proc(tr, sampling_rate=10):
    """
    Basic processing including downsampling, detrend, and demean.

    :param tr: raw trace
    :param sampling_rate:
    :return tr: trace after processing
    """
    # deep copy
    tr2 = tr.copy()
    tr2.interpolate(sampling_rate)
    tr2.detrend(type="linear")
    tr2.detrend(type="demean")
    return tr2


def _simple_proc(st, sampling_rate=10, njobs=1):
    """
    A parallel version of `_proc`, i.e., Basic processing including downsampling, detrend, and demean.

    :param st: an obspy stream
    :param sampling_rate: expected sampling rate
    :param njobs: number of jobs or CPU to use
    :return st: stream after processing
    """

    # downsampling, detrend, demean
    do_work = partial(_proc, sampling_rate=sampling_rate)

    # trace_list = []
    # for tr in st:
    #     trace_list.append(tr)
    #
    st2 = Stream()
    logging.info("simple processing for full event correlogram.")
    print("simple processing for full event correlogram.")
    if njobs == 1:
        logging.info('do work sequential (%d cores)', njobs)
        for tr in tqdm(st, total=len(st)):
            tr2 = do_work(tr)
            st2.append(tr2)
    else:
        logging.info('do work parallel (%d cores)', njobs)
        pool = multiprocessing.Pool(njobs)
        for tr2 in tqdm(pool.imap_unordered(do_work, st), total=len(st)):
            st2.append(tr2)
        pool.close()
        pool.join()

    return st2


def _proc_event_Z(file, **kwargs):
    """
    Processing a single component Z including downsampling, detrend, demean,
    remove instrumental response, selection of data by SNR>threshold,
    spectral whitening, temporal normalization and auto-correlation.

    For event type data.

    :param file:
    :param kwargs:
    :return:
    """

    try:
        # read file and return an obspy trace
        tr = read(file)[0]
        # process z-component only.
        if tr.stats.channel[-1] is not "Z":
            logging.warning("component is not Z: %s", file)
            return 0

        # first step is downsamping data to reduce computational burden
        sampling_rate = kwargs["preprocessing"]["sampling_rate"]
        tr.interpolate(sampling_rate)
        tr.detrend(type="linear")
        tr.detrend(type="demean")

        options = kwargs["data_selection_event"]
        snr_threshold = options["snr_threshold"]
        snr = _select_data_event(tr, **options)
        if snr < snr_threshold:
            return 0
        tr.stats.update({"snr": snr})
        logging.info("%s SNR is %f", file, snr)

        # remove instrumental response
        if kwargs["rm_resp_on"]:
            option_resp = kwargs["preprocessing"]["resp"]
            tr = remove_response(trace=tr, **option_resp)

        # spectral whitening
        if kwargs["whiten_on"]:
            option_whiten = kwargs["preprocessing"]["whiten"]
            tr = spectral_whitening(tr=tr, **option_whiten)

        # temporal normalization
        if kwargs["time_norm_on"]:
            option_time_norm = kwargs["preprocessing"]["time_norm"]
            tr = time_norm(tr, **option_time_norm)

        # autocorrelation
        options = kwargs["correlate"]
        tr = _autocorrelation(tr, **options)

        # write data to disk
        tr.stats.channel = "ZZ"  # change trace channel to 'ZZ' after autocorrelation
        trace_id = tr.id
        station_id = _get_station_id(tr)
        event_id = _get_event_id_tr(tr)
        filename = trace_id + "_" + event_id + ".pkl"
        outpath = kwargs["io"]["outpath"] + "/1_results"
        filepath = "/".join([outpath, station_id])
        # if not os.path.exists(filepath):
        try:
            os.makedirs(filepath)
        except:
            pass

        # extracting data, to have the same length of output data
        tlen = kwargs["correlate"]["window"]
        tlen = abs(tlen[1] - tlen[0])
        tr.trim(starttime=tr.stats.starttime, endtime=tr.stats.starttime + tlen, fill_value=0, pad=True)

        filen = filepath + "/" + filename
        force = kwargs["io"]["force"]
        if not force and os.path.exists(filen):
            pass
        else:
            tr.write(filen, format="PICKLE")

        return 1

    except:
        return 0


def _autocorrelation(tr, window=[-20, 70], filter=[0.5, 4], corners=2, zerophase=True):
    """
    Autocorrelation for event type data.

    :param tr:
    :param window:
    :param filter: A tuple or a list containing the lower and upper limit of filter.
    :param corners:
    :param zerophase:
    :return:

    .. Note::
        filter after autocorrelation.
    """
    tr2 = tr.copy()
    t1 = tr2.stats.onset + window[0]
    t2 = tr2.stats.onset + window[1]
    tr2.trim(starttime=t1, endtime=t2)
    npts = tr2.stats.npts
    data = tr2.data
    cc = correlate(in1=data, in2=data, mode="full")
    tr2.data = np.copy(cc)
    if filter is not None:
        tr2.filter(type="bandpass", freqmin=filter[0], freqmax=filter[1],
                   corners=corners, zerophase=zerophase)
    tr2.data = tr2.data[npts:]
    return tr2


def _crosscorrelation(tr1, tr2, window=[-20, 70], filter=[0.5, 4], corners=2, zerophase=True):
    """
    cross-correlation for event type data.

    :param tr1:
    :param tr2:
    :param window:
    :param filter:
    :param corners:
    :param zerophase:
    :return:
    """

    tra = tr1.copy()
    trb = tr2.copy()

    reverse = False
    if tra.stats.phase == "S":
        if tra.stats.channel == trb.stats.channel:
            # in this case, it is autocorrelation, which means SS
            window = window
        else:
            # in this case it is cross correlation over ZR/LQ or ZT/LT corresponding with Sp RF.
            # using different time window since converted P arrives earlier than the primary S.
            # if a window [-20, 70] is given, then a window [-70, 20] will be used in this case.
            w1, w2 = window
            window = [-w2, -w1]
            reverse = True

    t1 = tr2.stats.onset + window[0]
    t2 = tr2.stats.onset + window[1]
    tra.trim(starttime=t1, endtime=t2)
    trb.trim(starttime=t1, endtime=t2)
    npts = tra.stats.npts
    data = tra.data
    datb = trb.data

    cc = correlate(in1=data, in2=datb, mode="full")
    tr = tra.copy()
    tr.data = np.copy(cc)
    tr.filter(type="bandpass", freqmin=filter[0], freqmax=filter[1],
              corners=corners, zerophase=zerophase)
    if reverse:
        tr.data = np.copy(tr.data[::-1])
    tr.data = tr.data[npts:]
    return tr


def _select_data_event(tr, dist_range=[30, 90], magnitude=[5.0, 9.0],
                       snr_threshold=2.0, signal=[-10, 10], noise=[-100, -50],
                       waterlevel=1.0e-8):
    # select according to SNR
    distance = tr.stats.distance
    if distance < dist_range[0] or distance > dist_range[1]:
        return 0

    try:
        mag = tr.stats.magnitude
        if mag < magnitude[0] or mag > magnitude[1]:
            return 0
    except:
        pass

    # calculate SNR
    snr = _cal_event_snr(tr, signal=signal, noise=noise, waterlevel=waterlevel)

    return snr


def _cal_event_snr(tr, signal=[-10, 10], noise=[-100, -50], waterlevel=1.0e-8):
    """
    Calculation of SNR for event data.

    :param tr:
    :param signal:
    :param noise:
    :param waterlevel:
    :return:
    """

    tr_sig = tr.copy()
    tr_noi = tr.copy()

    t1 = tr.stats.onset + signal[0]
    t2 = tr.stats.onset + signal[1]
    if t1 < tr.stats.starttime or t2 > tr.stats.endtime:
        logging.warning("t1 < tr.stats.starttime")
        return 0
    tr_sig.trim(starttime=t1, endtime=t2)

    t1 = tr.stats.onset + noise[0]
    t2 = tr.stats.onset + noise[1]
    if t1 < tr.stats.starttime or t2 > tr.stats.endtime:
        logging.warning("t1 < tr.stats.starttime")
        return 0
    tr_noi.trim(starttime=t1, endtime=t2)

    sig = envelope(tr_sig.data)
    noi = envelope(tr_noi.data)
    snr = np.max(sig) / max(np.max(noi), waterlevel)

    return snr


def processing_noise_Z(jsonfile):
    """
    Parallel processing of noise autocorrelation.

    :param jsonfile:
    :return:
    """

    # load imported data from files
    kwargs = _load_json(jsonfile)
    njobs = kwargs["njobs"]
    datapath = kwargs["io"]["outpath"] + "/0_raw"

    logging.info("This may take a lot of time if dataset is huge.")

    # get filenames in Unix style for further processing
    files = glob.glob(datapath + "/*/*.pkl")
    logging.info("%d files found.", len(files))

    # processing includes downsampling, detrend, demean,
    # instrument response removal, spectral whitening, temporal normalization
    # autocorrelation and filter, then output results.
    do_work = partial(_proc_noise_Z, **kwargs)

    numbers = []
    if njobs == 1:
        logging.info('do work sequential (%d cores)', njobs)
        for file in tqdm(files, total=len(files)):
            num = do_work(file)
            numbers.append(num)
    else:
        logging.info('do work parallel (%d cores)', njobs)
        pool = multiprocessing.Pool(njobs)
        for num in tqdm(pool.imap_unordered(do_work, files), total=len(files)):
            numbers.append(num)
        pool.close()
        pool.join()

    logging.info("%d/%d files processed.", sum(numbers), len(files))


def _proc_noise_Z(file, **kwargs):
    """
    Internal functions of calculate z-comp autocorrelation for noise data.

    :param file:
    :param kwargs:
    :return:
    """

    # read file and return an obspy trace
    tr = read(file)[0]
    # process z-component only.
    if tr.stats.channel[-1] is not "Z":
        return 0

    # first step is downsamping data to reduce computational burden
    sampling_rate = kwargs["preprocessing"]["sampling_rate"]
    tr.interpolate(sampling_rate)
    tr.detrend(type="linear")
    tr.detrend(type="demean")

    # remove instrumental response
    if kwargs["rm_resp_on"]:
        option_resp = kwargs["preprocessing"]["resp"]
        tr = remove_response(trace=tr, **option_resp)

    # spectral whitening
    if kwargs["whiten_on"]:
        option_whiten = kwargs["preprocessing"]["whiten"]
        tr = spectral_whitening(tr=tr, **option_whiten)

    # temporal normalization
    if kwargs["time_norm_on"]:
        option_time_norm = kwargs["preprocessing"]["time_norm"]
        tr = time_norm(tr, **option_time_norm)

    # sliding autocorrelation
    options = kwargs["correlate_noise"]
    st = _sliding_autocorrelation(tr=tr, **options)

    # extracting data
    tlen = kwargs["correlate"]["window"]
    tlen = abs(tlen[1] - tlen[0])
    for tr in st:
        tr.trim(starttime=tr.stats.starttime, endtime=tr.stats.starttime + tlen, fill_value=0, pad=True)

    # write data to disk
    for tr in st:
        tr.stats.channel = "ZZ"  # change trace channel to 'ZZ' after autocorrelation
        trace_id = tr.id
        station_id = _get_station_id(tr)
        event_id = _get_noise_id(tr)
        filename = trace_id + "_" + event_id + ".pkl"
        outpath = kwargs["io"]["outpath"] + "/1_results"
        filepath = "/".join([outpath, station_id])
        # if not os.path.exists(filepath):
        try:
            os.makedirs(filepath)
        except:
            pass

        filen = filepath + "/" + filename
        force = kwargs["io"]["force"]
        if not force and os.path.exists(filen):
            pass
        else:
            tr.write(filen, format="PICKLE")

    return 1


def _sliding_autocorrelation(tr, length=3600, overlap=1800,
                             filter=[0.5, 4], corners=2, zerophase=True):
    """
    Sliding autocorrelation for noise data.

    :param tr:
    :param length:
    :param overlap:
    :param filter:
    :param corners:
    :param zerophase:
    :return:
    """
    trace = tr.copy()
    time_series = iter_time(tr=trace, length=length, overlap=overlap)
    # print(time_series)
    if len(time_series) < 1:
        return 0

    st = Stream()
    for t1, t2 in time_series:
        tr2 = trace.copy()
        tr2.trim(starttime=t1, endtime=t2)

        # get gaps
        gap_st = Stream([tr2])
        gaps = gap_st.get_gaps()
        if len(gaps) > 0:
            continue

        npts = tr2.stats.npts
        data = tr2.data
        cc = correlate(in1=data, in2=data, mode="full")
        tr2.data = np.copy(cc)
        tr2.filter(type="bandpass", freqmin=filter[0], freqmax=filter[1],
                   corners=corners, zerophase=zerophase)
        tr2.data = tr2.data[npts:]
        st.append(tr2)

    if len(st) < 1:
        return 0
    else:
        return st



# -*- coding: utf-8 -*-
"""
Plottings
"""
import matplotlib.pyplot as plt
from matplotlib.ticker import (FormatStrFormatter, MultipleLocator, AutoMinorLocator, FixedLocator, FixedFormatter, MaxNLocator)
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from matplotlib.markers import MarkerStyle
from matplotlib import cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

from acc.stack import linear_stack
from acc.profile import profile, get_profile_boxes
from acc.io import _load_json
from acc.migration import mig_one_station

import numpy as np
import warnings
from obspy import Stream
from obspy import read
from obspy.geodetics import gps2dist_azimuth

from collections import OrderedDict

import os
import glob
import math


def plot_acc_one_station(stream, fname=None, fig_width=7., trace_height=0.5,
                         stack_height=0.5, scale=2, scale_stack=10, fillcolors=("red", "blue"),
                         # trim=None,
                         info=(('back_azimuth', u'baz (°)', 'C0'),
                               ('distance', u'dist (°)', 'C3'))):
    """
    Plot auto- or correlogram for one station. Reproduced from the `rf` package.

    :param stream: stream to plot
    :param fname: filename to save plot to. Can be None. In this case
        the figure is left open.
    :param fig_width: width of figure in inches
    :param trace_height: height of one trace in inches
    :param stack_height: height of stack axes in inches
    :param scale: scale for individual traces
    :param fillcolors: fill colors for positive and negative wiggles
    :param info: Plot one additional axes showing maximal two entries of
        the stats object. Each entry in this list is a list consisting of
        three entries: key, label and color.
        info can be None. In this case no additional axes is plotted.
    """
    # :param trim: trim stream relative to onset before plotting using
    #      `~.rfstream.RFStream.slice2()`

    if len(stream) == 0:
        return

    stream.sort(keys=["slowness"])

    # if trim:
    #     stream = stream.slice2(*trim, reftime='onset')
    N = len(stream)
    # calculate lag times
    # stats = stream[0].stats
    # print(stats)
    # times = stream[0].times() - (stats.onset - stats.starttime)
    times = stream[0].times()
    # calculate axes and figure dimensions
    # big letters: inches, small letters: figure fraction
    H = trace_height
    HS = stack_height
    FB = 0.5
    FT = 0.2
    DW = 0.2
    FH = H * (N + 2) + HS + FB + FT + DW
    h = H / FH
    hs = HS / FH
    fb = FB / FH
    ft = FT / FH
    FL = 0.5 # figure left
    FR = 0.2 # figure right
    FW = fig_width # figure width
    FW3 = 0.8
    FW2 = FW - FL - FR - (DW + FW3) * bool(info)
    fl = FL / FW
    fr = FR / FW
    fw2 = FW2 / FW
    fw3 = FW3 / FW
    # init figure and axes
    fig = plt.figure(figsize=(FW, FH))
    ax1 = fig.add_axes([fl, fb, fw2, h * (N + 2)])
    if info:
        ax3 = fig.add_axes(
            [1 - fr - fw3, fb, fw3, h * (N + 2)], sharey=ax1)
        info = list(info)
        info[0] = [ax3] + list(info[0])
        if len(info) > 1:
            ax4 = ax3.twiny()
            info[1] = [ax4] + list(info[1])
    # plot individual receiver functions

    def _plot(ax, t, d, i):
        c1, c2 = fillcolors
        if c1:
            ax.fill_between(t, d + i, i, where=d >= 0, lw=0., facecolor=c1)
        if c2:
            ax.fill_between(t, d + i, i, where=d < 0, lw=0., facecolor=c2)
        ax.plot(t, d + i, 'k')
    # max_ = max(np.max(np.abs(tr.data)) for tr in stream)
    for i, tr in enumerate(stream):
        # _plot(ax1, times, tr.data / max_ * scale, i + 1)
        tr.normalize()
        _plot(ax1, times, tr.data * scale, i + 1)
    # plot right axes with header information
    for ax, header, label, color in info:
        data = [tr.stats[header] for tr in stream]
        ax.plot(data, 1 + np.arange(len(stream)), '.' + color, mec=color)
        ax.set_xlabel(label, color=color, size='small')
        if header == 'back_azimuth':
            ax.set_xticks(np.arange(5) * 90)
            ax.set_xticklabels(['0', '', '180', '', '360'], size='small')
        else:
            ax.xaxis.set_major_locator(MaxNLocator(4))
            for l in ax.get_xticklabels():
                l.set_fontsize('small')
        ax.xaxis.set_minor_locator(AutoMinorLocator())
    # set x and y limits
    ax1.set_xlim(times[0], times[-1])
    ax1.set_xlim(times[0], times[-1]/2)
    ax1.set_ylim(-0.5, N + 1.5)
    ax1.set_yticklabels('')
    ax1.set_xlabel('time (s)')
    ax1.xaxis.set_minor_locator(AutoMinorLocator())

    # plot stack
    # stack = stream.stack()
    trace = linear_stack(stream=stream, normalize=True)
    stack = Stream([trace])
    if len(stack) > 1:
        warnings.warn('Different stations or channels in one RF plot.')
    elif len(stack) == 1:
        stack.normalize()
        ax2 = fig.add_axes([fl, 1 - ft - hs, fw2, hs], sharex=ax1)
        _plot(ax2, times, stack[0].data * scale_stack, 0)
        for l in ax2.get_xticklabels():
            l.set_visible(False)
        ax2.yaxis.set_major_locator(MaxNLocator(4))
        for l in ax2.get_yticklabels():
            l.set_fontsize('small')
        # annotate plot with seed id
        bbox = dict(boxstyle='round', facecolor='white', alpha=0.8, lw=0)
        text = '%s traces  %s' % (len(stream), stack[0].id)
        ax2.annotate(text, (1 - 0.5 * fr, 1 - 0.5 * ft),
                     xycoords='figure fraction', va='top', ha='right',
                     bbox=bbox, clip_on=False)
        ax2.set_ylim(-1., 1.)
    # save plot
    if fname:
        fig.savefig(fname)
        plt.close(fig)
    else:
        return fig

def plot_ppoint(stream, fname="pp.pdf", depths=[30, 50, 80, 100, 150, 200]):

    colors = ["gray", "red", "blue", "orange", "green", "magenta", "cyan", "chocolate", "pink", "royalblue"]
    if len(depths) > 10:
        raise Exception("too many depths. Should be less than 10.")

    fig, ax = plt.subplots()

    for tr in stream:
        df = tr.stats.mig
        ax.scatter(tr.stats.station_longitude, tr.stats.station_latitude, c="black", marker="v", s=100)

        for i, depth in enumerate(depths):
            df2 = df[df["depth"] == depth]

            lat = df2["lat"].to_numpy()
            lon = df2["lon"].to_numpy()

            ax.scatter(lon, lat, c=colors[i], label=depth)

    handles, labels = plt.gca().get_legend_handles_labels()
    by_label = OrderedDict(zip(labels, handles))
    ax.legend(by_label.values(), by_label.keys(), title="Depth (km)")

    plt.tight_layout()
    plt.savefig(fname)



def plot_profile(jsonfile):

    kwargs = _load_json(jsonfile)
    paras = kwargs["profile"]

    latlon0 = paras["latlon0"]
    if latlon0 is None:
        # for a single station
        azimuth = 0
        dist = 400
    else:
        # for profile
        latlon1 = paras["latlon1"]
        lat1 = latlon0[0]
        lon1 = latlon0[1]
        lat2 = latlon1[0]
        lon2 = latlon1[1]
        dist, azimuth, baz = gps2dist_azimuth(lat1, lon1, lat2, lon2)
        dist /= 1000  # from m to km
        # padding 100 km on each side
        dist += 400

        lat0 = lat1 + math.cos((azimuth + 180)*math.pi/180) * 200 / 111.195
        lon0 = lon1 + math.sin((azimuth + 180)*math.pi/180) * 200 / 111.195
        latlon0 = (lat0, lon0)

    binsize = paras["binsize"]
    binwidth = paras["binwidth"]

    nbins = int(dist / binsize + 0.5) + 1
    dist = (nbins - 1) * binsize
    bins = np.linspace(0, dist, nbins)
    # print(bins)

    profile_id = paras["profile_id"]
    clip = paras["clip_percentage"] / 100.0

    path = kwargs["io"]["outpath"]

    depth_range = kwargs["plot"]["depth_range"]

    # each stations
    if latlon0 is None:
        files = glob.glob(path + "/migration_1station/*.pkl")
        files.sort()
        for file in files:
            stmig = read(file)
            lat0 = stmig[0].stats.station_latitude - 0.5 * (bins[-1] + bins[0]) / 111.195
            lon0 = stmig[0].stats.station_longitude
            latlon0 = (lat0, lon0)
            _plot_1station(stmig, latlon0, azimuth, bins, width=binwidth, clip=clip, savepath=path,
                           depth_range=depth_range)
    # profile
    else:
        wc = paras["wild_card"]
        files = glob.glob(path + "/migration_1station/*%s*.pkl" % wc)
        print("number of stations to stack: ", len(files))
        stmig = read(path + "/migration_1station/*%s*.pkl" % wc)
        _plot_stations(stmig, latlon0, azimuth, bins, width=binwidth, clip=clip, savepath=path,
                       depth_range=depth_range, profile_id=profile_id, **kwargs)


def _plot_1station(stmig, latlon0, azimuth, bins, width, clip, savepath, depth_range):
    # get boxes for mig-stacking
    boxes = get_profile_boxes(latlon0, azimuth=azimuth, bins=bins, width=width)

    # depth and stack array
    pos, depth, stack = profile(stream=stmig, boxes=boxes)

    # the setting makes the station location in the center of the image
    pos -= 0.5 * (bins[-1] - bins[0])

    dist_range = [-100, 100]

    # get station id
    tr = stmig[0]
    if tr.stats.location == "":
        station_id = ".".join([tr.stats.network, tr.stats.station])
    else:
        station_id = ".".join([tr.stats.network, tr.stats.station, tr.stats.location])
    print("Plotting - station id: ", station_id)

    # latitude corresponding to pos
    lats = []
    for b in boxes:
        # print(b["pos"], b["latlon"])
        lats.append(b["latlon"][0])

    extent = [np.min(pos), np.max(pos), np.min(depth), np.max(depth)]
    # extent = [np.min(lats), np.max(lats), np.min(depth), np.max(depth)]

    # normalization
    stack /= np.max(np.abs(stack))

    amp = np.sum(stack, axis=1)
    amp /= np.max(np.abs(amp))

    # print(extent)

    # return amp, stack, extent
    _plot(amp, depth, stack, extent, dist_range, depth_range, savepath, profile_id=station_id, clip=clip)


def _plot_stations(stmig, latlon0, azimuth, bins, width, clip, savepath, depth_range, profile_id, **kwargs):

    paras = kwargs["plot"]
    iclip = paras["image_scale"]
    wclip = paras["wavef_scale"]

    # get boxes for mig-stacking
    boxes = get_profile_boxes(latlon0, azimuth=azimuth, bins=bins, width=width)

    # depth and stack array
    pos, depth, stack = profile(stream=stmig, boxes=boxes)

    # the setting makes the station location in the center of the image
    # pos -= 0.5*(bins[-1]-bins[0])

    if paras["dist_range"] is None:
        dist_range = [np.min(pos), np.max(pos)]
    else:
        dist_range = paras["dist_range"]

    # get station id
    # tr = stmig[0]
    # if tr.stats.location == "":
    #     station_id = ".".join([tr.stats.network, tr.stats.station])
    # else:
    #     station_id = ".".join([tr.stats.network, tr.stats.station, tr.stats.location])
    print("Plotting - profile id: ", profile_id)

    # latitude corresponding to pos
    lats = []
    for b in boxes:
        # print(b["pos"], b["latlon"])
        lats.append(b["latlon"][0])

    extent = [np.min(pos), np.max(pos), np.min(depth), np.max(depth)]
    # extent = [np.min(lats), np.max(lats), np.min(depth), np.max(depth)]

    # normalization
    stack /= np.max(np.abs(stack))

    amp = np.sum(stack, axis=1)
    amp /= np.max(np.abs(amp))

    # print(extent)

    # return amp, stack, extent
    _plot(amp, depth, stack, extent, dist_range, depth_range, savepath, profile_id=profile_id, wclip=wclip, iclip=iclip)


def _plot(amp, depth, stack, extent, dist_range, depth_range, savepath, profile_id, wclip=1, iclip=1):
    # init figure and axes
    N = 256
    rdbu_r = cm.get_cmap('RdBu_r', N)
    newcolors = rdbu_r(np.linspace(0, 1, N))
    white = np.array([255 / 255, 255 / 255, 255 / 255, 1])
    a = int(N / 2 - N / 20)
    b = int(N / 2 + N / 20)
    newcolors[a:b, :] = white
    newcmp = ListedColormap(newcolors)

    fig, (ax2, ax1) = plt.subplots(nrows=1, ncols=2, figsize=(5.2, 3), sharey=True,
                                   gridspec_kw={'width_ratios': [0.6, 2]})
    im = ax1.imshow(stack*iclip, interpolation='bilinear', aspect="auto",
                    origin="lower", extent=extent,
                    cmap=newcmp, vmin=-1, vmax=1)

    ax1.yaxis.set_major_locator(MultipleLocator(50))
    ax1.yaxis.set_minor_locator(MultipleLocator(10))
    ax1.xaxis.set_minor_locator(MultipleLocator(50))
    ax1.grid(which="both", axis="y", linestyle=":", linewidth=0.5)

    ax1.annotate(profile_id,
                 xy=(0.995, 0.995), xycoords='axes fraction',
                 # xytext=(0.95, 0.05), textcoords=station_id,
                 # arrowprops=dict(facecolor='black', shrink=0.05),
                 horizontalalignment='right', verticalalignment='top')

    ax1.set_ylim(depth_range)
    ax1.set_xlim(dist_range)
    ax1.invert_yaxis()
    # ax1.invert_xaxis()

    ax2.plot(amp*wclip, depth, linewidth=0.75, color="k", alpha=0.7)
    ax2.fill_betweenx(depth, 0, amp*wclip, where=amp >= 0, facecolor='red', alpha=0.7)
    ax2.fill_betweenx(depth, 0, amp*wclip, where=amp < 0, facecolor='blue', alpha=0.7)
    ax2.grid(which="both", axis="y", linestyle=":", linewidth=0.5)
    ax2.set_xlim([-1, 1])
    ax2.set_xlabel("Amplitude")
    ax2.set_ylabel("Depth [km]")

    cbar = fig.colorbar(im, ax=ax1)
    cbar.set_label("Amplitude")

    ax1.set_xlabel("CRP Distance [km]")
    # ax1.set_ylabel("Depth [km]")

    plt.tight_layout()
    path = savepath + "/figures"
    if not os.path.exists(path):
        os.makedirs(path)
    filen = path + "/mig_crp_" + profile_id
    for fmt in [".pdf", ".png"]:
        fn = filen + fmt
        plt.savefig(fn)
    plt.close()





# -*- coding: utf-8 -*-
from obspy.taup import TauPyModel
import pandas as pd
from scipy.interpolate import interp1d
import numpy as np
from obspy.geodetics import gps2dist_azimuth
from acc.stack import pws_stack
from obspy import Stream

import os
import glob
from tqdm import tqdm
from functools import partial
from multiprocessing import Pool
import multiprocessing

from acc.io import _load_json
import logging
from obspy import read

# standalone version
# this function is supposed to be modified to parallel version.
def mig_one_station(stream, model="ak135", earth_radius=6378137.0, depth_range=(0, 300, 1)):

    try:
        # prior to the model used to calculate traveltime
        model = stream[0].stats.model
    except AttributeError:
        model = model

    taup_model = TauPyModel(model=model)
    for tr in stream:
        evdp = tr.stats.event_depth
        evla = tr.stats.event_latitude
        evlo = tr.stats.event_longitude

        stla = tr.stats.station_latitude
        stlo = tr.stats.station_longitude
        phase = tr.stats.phase

        # tr.filter(type="bandpass", freqmin=0.05, freqmax=0.5, zerophase=True)

        arrivals = taup_model.get_ray_paths_geo(source_depth_in_km=evdp,
                                                source_latitude_in_deg=evla, source_longitude_in_deg=evlo,
                                                receiver_latitude_in_deg=stla, receiver_longitude_in_deg=stlo,
                                                phase_list=(phase,), resample=True)

        arr = arrivals[0]
        # raypath coordinates. The dataframe contains six columns, ray_p, traveltime, dist in rad, depth, lat and lon.
        df = pd.DataFrame(arr.path)

        # one-way reflection time
        time = tr.times() / 2
        data = tr.data

        # ray path information: traveltime, depth, and coordinates
        # reverse time: the maximum traveltime minus traveltime at varied depths or points
        df["time_reverse"] = df["time"].iloc[-1] - df["time"]
        # get sub-dataset with traveltime (ray path) slightly longer than the one-way reflected traveltime from acc
        # for convenience of interpolation
        df2 = df[df["time_reverse"] < time[-1] * 1.2]

        # convert the raypath information from dataframe to numpy array
        ttime = df["time_reverse"].to_numpy()
        depth = df["depth"].to_numpy()
        lat = df["lat"].to_numpy()
        lon = df["lon"].to_numpy()

        # simple migration by back projection
        fdepth = interp1d(ttime, depth)
        flat = interp1d(ttime, lat)
        flon = interp1d(ttime, lon)

        # interpolation fitting in with one-way reflection time to accomplish back projection
        depths = fdepth(time)
        lats = flat(time)
        lons = flon(time)

        # calculate the distances from pierce points to station at different depths
        # or varied radius
        dists = np.zeros_like(depths)
        for i, d in enumerate(depths):
            dists[i], az, baz = gps2dist_azimuth(lat1=tr.stats.station_latitude, lon1=tr.stats.station_longitude,
                                                 lat2=lats[i], lon2=lons[i], a=earth_radius-d)
            # convert the unit from m to km. If the piere point is to the south of the station then distance is negative.
            if lats[i] <= tr.stats.station_latitude:
                dists[i] = -dists[i]
        dists /= 1000

        # the following several lines can be commented. This was firstly written to save data with
        # irregular depth sampling intervals.
        mig1sta = pd.DataFrame(columns=["time", "depth", "dist", "lat", "lon", "data"])
        mig1sta["depth"] = depths
        mig1sta["lat"] = lats
        mig1sta["lon"] = lons
        mig1sta["time"] = time
        tr.normalize()
        mig1sta["data"] = tr.data
        mig1sta["dist"] = dists

        # second interpolation to regular depth grid. after back-projection the depth sampling is irregular.
        # depth range 0-300 km with an interval of 0.5 km.
        delta = depth_range[2]
        d = np.arange(depth_range[0], depth_range[1]+delta, delta)
        tr.stats.delta = delta
        ftime = interp1d(depths, time)
        fdist = interp1d(depths, dists)
        flat = interp1d(depths, lats)
        flon = interp1d(depths, lons)
        fdata = interp1d(depths, tr.data)

        time = ftime(d)
        dists = fdist(d)
        lats = flat(d)
        lons = flon(d)
        tr.data = fdata(d)
        depths = np.copy(d)

        mig1sta = pd.DataFrame(columns=["time", "depth", "dist", "lat", "lon", "data"])
        mig1sta["depth"] = depths
        mig1sta["lat"] = lats
        mig1sta["lon"] = lons
        mig1sta["time"] = time
        tr.normalize()
        mig1sta["data"] = tr.data
        mig1sta["dist"] = dists

        header = {"path": df2, "mig":mig1sta}
        tr.stats.update(header)

    return stream


def _mig_1(path, model="ak135"):
    # read autocorrelograms of a given station saved in `path`
    st = read(path + "/*pkl")
    st_mig = mig_one_station(stream=st, model=model, earth_radius=6378137.0)

    return st_mig


def migration_1station(jsonfile):

    kwargs = _load_json(jsonfile)
    io = kwargs["io"]

    njobs = kwargs["njobs"]
    if njobs > multiprocessing.cpu_count():
        njobs = multiprocessing.cpu_count()

    # datapath containing auto-correlograms
    path = io["outpath"] + "/1_results"
    temp = glob.glob(path + "/*")
    stations = []
    for t in temp:
        if os.path.isdir(t):
            stations.append(t)

    model = kwargs["migration"]["model"]
    if model is None:
        model = kwargs["tt_model"]

    do_work = partial(_mig_1, model=model)

    st_mig_stations = []
    if njobs == 1:
        logging.info('do work sequential (%d cores)', njobs)
        for sta in tqdm(stations, total=len(stations)):
            st = do_work(sta)
            st_mig_stations.append(st)
    else:
        logging.debug('do work parallel (%d cores)', njobs)
        pool = multiprocessing.Pool(njobs)
        for st in tqdm(pool.imap_unordered(do_work, stations), total=len(stations)):
            st_mig_stations.append(st)
        pool.close()
        pool.join()

    # save data into disk
    outpath = io["outpath"] + "/migration_1station"
    # if not os.path.exists(outpath):
    try:
        os.makedirs(outpath)
    except:
        pass
    for st in st_mig_stations:
        # station id
        tr = st[0]
        if tr.stats.location == "":
            station_id = ".".join([tr.stats.network, tr.stats.station])
        else:
            station_id = ".".join([tr.stats.network, tr.stats.station, tr.stats.location])
        fn = outpath + "/" + station_id + ".pkl"
        st.write(fn, format="PICKLE")


def migration_one_station_stack(stream, method="PWS", power=2, time_range=[8, 16], coeff=0.5):
    """
    Stacking of migrated traces for one station. Currently not used.

    :param stream:
    :param method: "PWS" for phase-weighted stacking and "linear" for linear stacking
    :param power: used by "PWS" only. power=0 is linear stacking
    :param time_range: if None do simple stacking (PWS or linear). if a list contains two elements then
    it will select traces with high resemblance with the initial stacking to get final stacked trace.
    :param coeff: traces with correlation efficient higher than the value will be selected.
    :return: trace after stacking

    .. Note::
        The one more procedure is to improve signal-to-noise ratio. You may check the stats header `corrstack`.
        If the key equals zero, the stacking is the general ones, otherwise the correlated stacking are implemented.
        The value denotes the number of stacked traces.
    """
    # first initial stacking and find similar traces for final stacking.
    tr_stk = pws_stack(stream, power=power, normalize=True)

    h = {"corrstack": 0}
    tr_stk.stats.update(h)

    if time_range is None:
        return tr_stk

    delta = tr_stk.stats.delta
    i1 = int(time_range[0] / delta)
    i2 = int(time_range[1] / delta)

    data1 = np.copy(tr_stk.data[i1:i2])

    traces = []
    for tr in stream:
        data2 = np.copy(tr.data[i1:i2])
        c = np.corrcoef(data1, data2)
        if c[1][0] >= coeff:
            traces.append(tr)

    if len(traces) < 1:
        return tr_stk

    # final stacking with high coherence
    st2 = Stream(traces=traces)
    tr_stk = pws_stack(st2, power=power, normalize=True)
    h = {"corrstack": len(traces)}
    tr_stk.stats.update(h)

    return tr_stk








# -*- coding: utf-8 -*-
# Copyright 2019- Weijia Sun, MIT license

"""
stacking
"""
import os
import glob
import numpy as np
from scipy.signal import hilbert
import commentjson
from obspy import read, Stream

from acc.io import _load_json

from tqdm import tqdm
from functools import partial
from multiprocessing import Pool
import multiprocessing
import logging

logging.basicConfig(level=logging.NOTSET, format='%(asctime)s %(filename)s[line:%(lineno)d] %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                    filename="acc.processing.log", filemode="a"
                    )


def linear_stack(stream, normalize=True):
    if normalize:
        stream.normalize(global_max=False)

    stk = np.zeros_like(stream[0].data)
    for tr in stream:
        stk += tr.data

    stk /= len(stream)

    tr = stream[0]
    tr.data = np.copy(stk)
    tr.stats.starttime = 0
    tr.stats.update({"number_of_ray": len(stream), "stack_method":{"linear"}})

    return tr


def pws_stack(stream, power=2, normalize=True):
    if normalize:
        stream.normalize(global_max=False)

    # safe guard to be a positive integer number
    power = int(abs(power))

    tr_lin = linear_stack(stream, normalize)
    if power == 0:
        return tr_lin

    phase = np.zeros_like(stream[0].data)
    for tr in stream:
        # hilbert transform
        hilb = hilbert(tr.data)
        phase += np.abs(hilb / np.abs(hilb))

    phase /= len(stream)
    power = int(power)
    stk = tr_lin.data * np.power(phase, power)

    tr = stream[0]
    tr.data = np.copy(stk)
    tr.stats.starttime = 0
    tr.stats.update({"number_of_ray": len(stream), "stack_method":{"PWS", power}})

    return tr


def stack(jsonfile):
    kwargs = _load_json(jsonfile)

    path = kwargs["io"]["outpath"]
    stations = glob.glob(path + "/1_results/*")
    njobs = kwargs["njobs"]

    do_work = partial(_stack, **kwargs)

    numbers = []
    if njobs == 1:
        logging.info('do work sequential (%d cores)', njobs)
        for sta in tqdm(stations, total=len(stations)):
            num = do_work(sta)
            numbers.append(num)
    else:
        logging.info('do work parallel (%d cores)', njobs)
        pool = multiprocessing.Pool(njobs)
        for num in tqdm(pool.imap_unordered(do_work, stations), total=len(stations)):
            numbers.append(num)
        pool.close()
        pool.join()


def _stack(path, **kwargs):

    files = glob.glob(path + "/*")
    # stations = glob.glob(path + "/*")
    # print(stations)

    # for sta in stations:
    #     files = glob.glob(sta + "/*")

    st = Stream()
    for file in files:
        try:
            tr = read(file)[0]
        except:
            continue
        st.append(tr)

    # get channels
    channels = []
    for tr in st:
        chn = tr.stats.channel
        if chn not in channels:
            channels.append(chn)

    method = kwargs["stack"]["method"]
    power = kwargs["stack"]["power"]
    outpath = kwargs["io"]["outpath"] + "/1a_stack"
    try:
        os.makedirs(outpath)
    except:
        pass

    stream = st.copy()
    for chn in channels:

        st = stream.select(channel=chn)

        if method == "linear":
            tr = linear_stack(st, normalize=True)
        elif method == "PWS":
            tr = pws_stack(st, power, normalize=True)
        elif method == "bootstrap_linear" or method == "bootstrap_PWS":
            # note tr here is a stream containing two traces, mean and std
            tr = _bootstrap(st, normalize=True, **kwargs)

        filen = outpath + "/" + tr.id + "_%s.pkl" % method
        tr.write(filename=filen, format="PICKLE")

    return 0

def _bootstrap(st, normalize=True, **kwargs):

    if normalize:
        stream.normalize(global_max=False)

    # generate randoma number
    ntrace = len(st)
    options = kwargs["stack"]
    n_iter = options["n_iter"]
    percentage = options["percentage"]
    seed = kwargs["seed"]
    rand_list = _gen_random_number(low=0, high=ntrace,
                                   n_iter=n_iter, percentage=percentage, seed=seed)

    # stacking
    method = kwargs["stack"]["method"]
    power = kwargs["stack"]["power"]

    # method is bootstrap_linear or bootstrap_PWS
    method = method.split("_")[1]

    npts = st[0].stats.npts
    boot = np.zeros([n_iter, npts])
    m = 0
    for r in rand_list:
        # random stream
        st_rand = Stream()
        for i in r:
            tr = st[i]
            st_rand.append(tr)

        # stack
        if method == "linear":
            tr_stk = linear_stack(st_rand, normalize=True)
        elif method == "PWS":
            tr_stk = pws_stack(st_rand, power, normalize=True)

        boot[i] = np.copy(tr_stk.data)

    # mean value
    tr_mean = st[0].copy()
    tr_std = st[0].copy()
    tr_mean.data = np.mean(boot, axis=0)
    tr_mean.stats.update({"statistics":"mean"})
    tr_std.data = np.std(boot, axis=0)
    tr_st.stats.update({"statistics":"std"})

    st2 = Stream()
    st2.append(tr_mean)
    st2.append(tr_std)
    return st2


def _gen_random_number(low=0, high=100, n_iter=100, percentage=0.9, seed=59):

    np.random.seed(seed)

    n_trace = int(percentage * high)

    rand_list = []
    for i in range(n_iter):
        r = np.random.randint(low=high, high=high)
        rand_list.append(r)

        # reset random seed. use the multiplication of the first nonzero value
        # generated random number from the previous random seed as the new seed
        s = 1
        counter = 0
        for k in r:
            if k != 0:
                s *= k
            if counter > 5:
                break
            counter += 1
        np.random.seed(seed=s)

    return rand_list








# -*- coding: utf-8 -*-
import click

from acc.io import import_data
from acc.processing import processing_event_Z
from acc.processing import processing_noise_Z
from acc.migration import migration_1station
from acc.plotting import plot_profile

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
def run():
    pass


@run.command()
@click.option("-p", "--paras", default="conf.json", help="The parameter JSON file")
def importdata(**kwargs):
    """Import data and convert to internal format."""
    import_data(kwargs["paras"])


@run.command()
@click.option("-p", "--paras", default="conf.json", help="The parameter JSON file")
def calevent(**kwargs):
    """Calculate teleseismic P-wave autocorrelograms."""
    processing_event_Z(kwargs["paras"])


@run.command()
@click.option("-p", "--paras", default="conf.json", help="The parameter JSON file")
def calnoise(**kwargs):
    """Calculate ambient noise autocorrelograms."""
    processing_noise_Z(kwargs["paras"])


@run.command()
@click.option("-p", "--paras", default="conf.json", help="The parameter JSON file")
def migration(**kwargs):
    """Migrate teleseismic P-wave autocorrelograms."""
    migration_1station(kwargs["paras"])


@run.command()
@click.option("-p", "--paras", default="conf.json", help="The parameter JSON file")
def plotprofile(**kwargs):
    """Plot migrated P-wave profile."""
    plot_profile(kwargs["paras"])


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
"""
I/O modules
"""
import os
import commentjson
import glob
from obspy import read
from obspy.taup import TauPyModel
from obspy.geodetics import gps2dist_azimuth
from obspy.geodetics import kilometers2degrees

from tqdm import tqdm
from functools import partial
from multiprocessing import Pool
import multiprocessing

import logging

logging.basicConfig(level=logging.NOTSET, format='%(asctime)s %(filename)s[line:%(lineno)d] %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                    filename="acc.io.log", filemode="a"
                    )


def _load_json(jsonfile):
    """
    Load parameters from the json formatted file

    :param str jsonfile: json file containing parameters
    :return dict kwargs: dictionary containing parameters
    """
    with open(jsonfile, "r") as fp:
        kwargs = commentjson.load(fp)

    if kwargs["njobs"] is None:
        kwargs["njobs"] = multiprocessing.cpu_count()

    return kwargs


def _acc_read(file, outpath, acc_type, phase, force=True, tt_model="ak135", depth_unit="km"):
    """
    Read seismic data.

    :param file: file name to be read
    :param outpath: where to save the data
    :param acc_type: event or noise
    :param phase: if acc_type='event', then calculate the traveltime of the phase
    :param force: force to overwrite when file exists if force is True
    :param tt_model: 1-d theoretical model, e.g., ak135 used to calculate traveltime of a specific phase
    :param depth_unit: unit of depth, general one is km, but in some cases it could be m.
    :return:
    """

    tr = read(file)[0]

    station_id = _get_station_id(tr)
    trace_id = tr.id

    valid = 0
    if acc_type == "event":
        # tt_model = data_selection["tt_model"]
        event_id = _get_event_id(tr)
        # depth_unit = data_selection["depth_unit"]
        tr = _get_event_data(tr, tt_model, phase=phase, acc_type=acc_type, depth_unit=depth_unit)
        if tr is None:
            logging.warn("%s no valid arrival for phase %s", file, phase)
            return 0
        else:
            valid = 1
    elif acc_type == "noise":
        event_id = _get_noise_id(tr)
        tr = _get_noise_data(tr, acc_type=acc_type)
        valid = 1

    filename = trace_id + "_" + event_id + ".pkl"
    filepath = "/".join([outpath, station_id])
    # if not os.path.exists(filepath):
    try:
        os.makedirs(filepath)
    except:
        pass

    filen = filepath + "/" + filename
    if not force and os.path.exists(filen):
        pass
    else:
        tr.write(filen, format="PICKLE")
    return valid


def _get_sac_origin(tr):
    """
    Get the origin time of an event trace in sac format.

    :param tr: A event trace
    :return origin: origin time of an event

    .. Note::
        The trace should be sac formatted.

    """
    try:
        origin = tr.stats.starttime - tr.stats.sac.b + tr.stats.sac.o
    except AttributeError:
        logging.critical("no valid event origin found in header: tr.stats.sac.o")
        logging.critical("Please check acc_type='event' or 'noise'")
        raise AttributeError
    return origin


def _get_event_id(tr):
    """
    Get event id from a sac-formatted trace.

    :param tr:
    :return str event_id: event id
    """

    origin = _get_sac_origin(tr)
    event_id = origin.datetime.strftime("%Y%m%d%H%M%S")
    return event_id


def _get_event_id_tr(tr):
    """
    Get event id from a obspy trace. The obspy trace should have the event_time keyword.

    :param tr: a obspy trace
    :return str event_id: event id in "%Y%m%d%H%M%S", e.g., '20191019200909'
    """
    origin = tr.stats.event_time
    event_id = origin.datetime.strftime("%Y%m%d%H%M%S")
    return event_id


def _get_noise_id(tr):
    """
    Get trace id for noise type data. e.g., 'starttime-endtime'

    :param tr: an obspy trace
    :return str event_id: noise id
    """
    starttime = tr.stats.starttime
    endtime = tr.stats.endtime
    event_id = "-".join([starttime.strftime("%Y%m%d%H%M%S"), endtime.strftime("%Y%m%d%H%M%S")])
    return event_id


def _get_station_id(tr):
    """
    Get station id of a given trace.

    :param tr: trace
    :return station_id: station id formatted as '{newwork}.{station}.{location}'.
    """

    s = tr.id
    s = s.split(".")
    if s[2] == "":
        station_id = ".".join(s[:2])
    else:
        station_id = ".".join(s[:3])

    return station_id


def _get_noise_data(tr, acc_type):
    """
    Get update sac-trace header to obspy trace header station_latitude etc.

    :param tr:
    :param acc_type:
    :return tr:
    """

    station_longitude = tr.stats.sac.stlo
    station_latitude = tr.stats.sac.stla
    station_elevation = tr.stats.sac.stel

    header = {"type": acc_type,
              "station_latitude": station_latitude, "station_longitude": station_longitude,
              "station_elevation": station_elevation
              }

    tr.stats.update(header)

    return tr


def _get_event_data(tr, tt_model, phase, acc_type, depth_unit="km"):
    """
    Update a sac trace to a obspy trace and update trace header,
    and calculate theoretical traveltime of a specific model and phase

    :param tr:
    :param tt_model:
    :param phase:
    :param acc_type:
    :param depth_unit:
    :return:

    .. Note::
        The input trace should be read from sac-formatted files.

        depth_unit is not used. if depth>1000 then unit should be meter,
        since no events deeper than 700 km on the earth.

    """

    model = TauPyModel(model=tt_model)

    event_longitude = tr.stats.sac.evlo
    event_latitude = tr.stats.sac.evla
    event_depth = tr.stats.sac.evdp
    event_magnitude = tr.stats.sac.mag

    # if depth_unit == "m":
    #     event_depth /= 1000.0
    # in this case, the depth_unit is considered to be m.
    if event_depth > 1000:
        event_depth /= 1000

    station_longitude = tr.stats.sac.stlo
    station_latitude = tr.stats.sac.stla
    station_elevation = tr.stats.sac.stel

    try:
        component_azimuth = tr.stats.sac.cmpaz
        component_inclination = tr.stats.sac.cmpinc
    except:
        print(tr.stats)
        if tr.stats.channel[-1] == "Z":
            component_azimuth = 0
            component_inclination = 0
        elif tr.stats.channel[-1] == "N":
            component_azimuth = 0
            component_inclination = 90
        elif tr.stats.channel[-1] == "E":
            component_azimuth = 90
            component_inclination = 90



    event_time = _get_sac_origin(tr)

    distance, azimuth, back_azimuth = gps2dist_azimuth(lat1=event_latitude, lon1=event_longitude,
                                                       lat2=station_latitude, lon2=station_longitude,
                                                       a=6378137.0, f=0.0033528106647474805)
    distance = kilometers2degrees(kilometer=distance / 1000.0)

    # travel time, slowness, inclinations
    arrivals = model.get_travel_times(source_depth_in_km=event_depth,
                                      distance_in_degree=distance,
                                      phase_list=[phase])
    if len(arrivals) < 1:
        return None

    arr = arrivals[0]

    onset = event_time + arr.time
    phase = phase
    inclination = arr.incident_angle
    slowness = arr.ray_param

    # pierce points
    # pp_latitude
    # pp_longitude
    # pp_depth

    # ray paths
    # arrivals = model.get_travel_times(source_depth_in_km=event_depth,
    #                                   distance_in_degree=distance,
    #                                   phase_list=[phase])

    header = {"model": tt_model, "type": acc_type,
              "event_latitude": event_latitude, "event_longitude": event_longitude, "event_depth": event_depth,
              "event_time": event_time, "event_magnitude": event_magnitude,
              "station_latitude": station_latitude, "station_longitude": station_longitude,
              "station_elevation": station_elevation,
              "component_azimuth": component_azimuth, "component_inclination":component_inclination,
              "onset": onset, "phase": phase, "inclination": inclination, "slowness": slowness,
              "distance": distance, "azimuth": azimuth, "back_azimuth": back_azimuth
              }

    tr.stats.update(header)

    return tr


def import_data(jsonfile):
    """
    Import data from external media

    :param jsonfile: parameter filename
    :return:
    """

    kwargs = _load_json(jsonfile)
    io = kwargs["io"]
    # data_selection = kwargs["data_selection"]
    njobs = kwargs["njobs"]

    if njobs > multiprocessing.cpu_count():
        njobs = multiprocessing.cpu_count()

    files = glob.glob(io["data"])
    outpath = io["outpath"] + "/0_raw"
    acc_type = kwargs["acc_type"]
    phase = kwargs["phase"]
    force = io["force"]
    tt_model = kwargs["tt_model"]
    depth_unit = kwargs["depth_unit"]

    if acc_type not in ["event", 'noise']:
        print("Invalid acc_type: %s. Aborted." % acc_type)
        logging.error("Invalid acc_type: %s. Aborted.", acc_type)
        exit(-1)

    do_work = partial(_acc_read, outpath=outpath, acc_type=acc_type,
                      phase=phase, force=force, tt_model=tt_model, depth_unit=depth_unit)
    numbers = []
    if njobs == 1:
        logging.info('do work sequential (%d cores)', njobs)
        for file in tqdm(files, total=len(files)):
            num = do_work(file)
            numbers.append(num)
    else:
        logging.debug('do work parallel (%d cores)', njobs)
        pool = multiprocessing.Pool(njobs)
        for num in tqdm(pool.imap_unordered(do_work, files), total=len(files)):
            numbers.append(num)
        pool.close()
        pool.join()

    logging.info("%d/%d files imported.", sum(numbers), len(files))



# Copyright 2019 Weijia Sun, MIT license
import os.path
import re

from setuptools import find_packages, setup


def find_version(*paths):
    fname = os.path.join(os.path.dirname(__file__), *paths)
    with open(fname) as fp:
        code = fp.read()
    match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", code, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string.")


VERSION = find_version('acc', '__init__.py')
DESCRIPTION = 'Auto-Correlogram Calculation in seismology'
# LONG_DESCRIPTION = (
#     'Please look at the project site for tutorials and information.')
with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()


ENTRY_POINTS = {
    'console_scripts': ['acc=acc.main:run',
                        ]}

REQUIRES = ['cartopy', "click", 'commentjson', 'geographiclib',
            'matplotlib>=2', 'numpy',
            'obspy>=1.0.3', "pandas",
            'setuptools', 'shapely', 'scipy>=0.19.0', 'tqdm'
            ]

# EXTRAS_REQUIRE = {
#     'doc': ['sphinx', 'alabaster'],  # and decorator, obspy
#     'h5': ['obspyh5>=0.3']}

CLASSIFIERS = [
    'Environment :: Console',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Topic :: Scientific/Engineering :: Physics'
]


setup(name='seis-acc',
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type="text/markdown",
      url='https://github.com/weijias-opensource/acc',
      author='Weijia SUN',
      author_email='weijia_sun@163.com',
      license='MIT',
      packages=find_packages(),
      package_dir={'acc': 'acc'},
      install_requires=REQUIRES,
      # extras_require=EXTRAS_REQUIRE,
      entry_points=ENTRY_POINTS,
      # please note the entry_points
      # The magic is in the entry_points parameter. Below console_scripts, each line identifies one console script.
      # The first part before the equals sign (=) is the name of the script that should be generated,
      # the second part is the import path followed by a colon (:) with the Click command.
      # entry_points={"console_scripts": ['acc=acc.main:run',],},
      include_package_data=True,
      zip_safe=False,
      classifiers=CLASSIFIERS
      )



# -*- coding: utf-8 -*-
# Copyright 2019- Weijia Sun, MIT license
"""Preprocessing and correlation"""
import glob
# from functools import partial
# import itertools
import logging
# import multiprocessing

import numpy as np

from scipy.signal import freqz, iirfilter
from scipy.fftpack import fft, ifft, fftshift, ifftshift, next_fast_len

from acc.util import smooth as smooth_func
from acc.util import IterMultipleComponents

from obspy import Stream
from obspy import read_inventory
from obspy.signal.rotate import rotate2zne

log = logging.getLogger('acc.core.log')


def spectral_whitening(tr, smooth=None, filter=None,
                       waterlevel=1e-8, corners=2, zerophase=True):
    """
    Apply spectral whitening to data

    Data is divided by its smoothed (Default: None) amplitude spectrum.

    :param tr: trace to manipulate
    :param smooth: length of smoothing window in Hz
        (default None -> no smoothing)
    :param filter: filter spectrum with bandpass after whitening
        (tuple with min and max frequency)
        (default None -> no filter)
    :param waterlevel: waterlevel relative to mean of spectrum
    :param mask_again: weather to mask array after this operation again and
        set the corresponding data to 0
    :param corners: parameters parsing to filter,
    :param zerophase: parameters parsing to filter

    :return: whitened data
    """

    sr = tr.stats.sampling_rate
    data = np.copy(tr.data)
    # data = _fill_array(data, fill_value=0)
    # mask = np.ma.getmask(data)

    # transform to frequency domain
    nfft = next_fast_len(len(data))
    spec = fft(data, nfft)

    # amplitude spectrum
    spec_ampl = np.abs(spec)

    # normalization
    spec_ampl /= np.max(spec_ampl)
    spec_ampl_raw = np.copy(spec_ampl)

    # smooth
    if smooth:
        smooth = int(smooth * nfft / sr)
        spec_ampl = ifftshift(smooth_func(fftshift(spec_ampl), smooth))
        spec_ampl_smth = np.copy(spec_ampl)

    # save guard against division by 0
    spec_ampl[spec_ampl < waterlevel] = waterlevel

    # make the spectrum have the equivalent amplitude before/after smooth
    if smooth:
        scale = np.max(spec_ampl_raw) / np.max(spec_ampl_smth)
        spec /= spec_ampl * scale
    else:
        spec /= spec_ampl

    # FFT back to time domain
    ret = np.real(ifft(spec, nfft)[:len(data)])
    tr.data = ret

    # filter
    if filter is not None:
        tr.filter(type="bandpass", freqmin=filter[0], freqmax=filter[1],
                  corners=corners, zerophase=zerophase)

    return tr


def _fill_array(data, mask=None, fill_value=None):
    """
    Mask numpy array and/or fill array value without demasking.

    Additionally set fill_value to value.
    If data is not a MaskedArray and mask is None returns silently data.

    :param mask: apply mask to array
    :param fill_value: fill value
    """
    if mask is not None and mask is not False:
        data = np.ma.MaskedArray(data, mask=mask, copy=False)
    if np.ma.is_masked(data) and fill_value is not None:
        data._data[data.mask] = fill_value
        np.ma.set_fill_value(data, fill_value)
    #    elif not np.ma.is_masked(data):
    #        data = np.ma.filled(data)
    return data


def time_norm(tr, method, time_length=5, filter=(0.01, 1),
              corners=2, zerophase=True, waterlevel=1.0e-8):
    """
    Calculate normalized data, see e.g. Bensen et al. (2007, GJI)

    :param tr: Trace to manipulate
    :param str method:
        1bit: reduce data to +1 if >0 and -1 if <0\n
        run_abs_mean: running absolute mean normalization
    :param float time_length: time length to be smoothed, default 5 sec
    :param tuple filter: bandpass frequency band, default (0.01, 1) Hz
    :param int corners: corners default 2
    :param bool zerophase: default True
    :param float waterlevel: waterlevel to safe guard division

    :return: normalized data
    """

    data = tr.data
    data = _fill_array(data, fill_value=0)
    mask = np.ma.getmask(data)

    if method == '1bit':
        np.sign(data, out=data)

    elif method == 'run_abs_mean':
        data = _run_abs_mean(tr, time_length=time_length, filter=filter,
                             corners=corners, zerophase=zerophase,
                             waterlevel=waterlevel)
    else:
        msg = 'The method passed to time_norm is not known: %s.' % method
        raise ValueError(msg)

    tr.data = _fill_array(data, mask=mask, fill_value=0)
    return tr


def _run_abs_mean(tr, time_length=5, filter=(0.01, 1),
                  corners=4, zerophase=True, waterlevel=1e-8):
    """
    running absolute mean normalization

    :param tr: trace to be normalized
    :param float time_length: time length to be smoothed, default 5 sec
    :param tuple filter: bandpass frequency band, default (0.01, 1) Hz
    :param int corners: corners default 4
    :param bool zerophase: default True
    :param float waterlevel: waterlevel to safe guard division

    :return: data after temporal normalization
    """

    tr2 = tr.copy()
    tr2.filter(type="bandpass", freqmin=filter[0], freqmax=filter[1],
               corners=corners, zerophase=zerophase)
    data = np.abs(tr2.data)

    nl = int(time_length / tr2.stats.delta)
    smth_data = smooth_func(x=data, window_len=nl,
                            window='flat', method='zeros')

    # save guard against division by 0
    smth_data[smth_data < waterlevel] = waterlevel

    data = np.copy(tr.data)
    data /= smth_data

    return data


def rotation(stream, method="NE->RT", acc_type="event"):
    """
    Rotatation.

    :param stream: obspy stream to be rotated
    :param method: "NE->RT" or "ZNE->LQT"
    :param acc_type: "event", "noise"
    :return stream: obspy stream after rotation.

    .. note::
        The keywords `component_azimuth` and `component_inclination` must be given in the stats.
        Currently, only acc_type="event" are well debugged.
        The input 3-component data should be in the roughly same periods or time window.
    """

    rot_stream = Stream()

    # function to extract 3-component stream.
    # if the data are given event the data, then use onset as the key word to find the 3-component data
    if acc_type == "event":
        key = "onset"
    elif acc_type == "noise":
        key = "starttime"

    def iter3c(stream):
        return IterMultipleComponents(stream, key=key, number_components=(2, 3))

    for st3c in iter3c(stream):

        tmin, tmax = _find_start_end_time(stream=st3c)
        if tmin > tmax:
            continue
        st3c.trim(starttime=tmin, endtime=tmax)

        # handling the borehole components 1, 2, Z or 1, 2, 3.
        comp = []
        for tr in st3c:
            comp.append(tr.stats.channel[-1])
        comp.sort()
        cc = "".join(comp)
        if cc == "12Z":
            cc = "Z12"
        # rotate2zne
        if cc in ["123", "Z12"]:
            zne = rotate2zne(st3c[0].data, st3c[0].stats.component_azimuth, st3c[0].stats.component_inclination,
                             st3c[1].data, st3c[1].stats.component_azimuth, st3c[1].stats.component_inclination,
                             st3c[2].data, st3c[2].stats.component_azimuth, st3c[2].stats.component_inclination
                             )

            for tr, new_data, component in zip(st3c, zne, "ZNE"):
                tr.data = new_data
                tr.stats.channel = tr.stats.channel[:-1] + component
        # rotate to various coordinates
        st3c.rotate(method=method)
        rot_stream += st3c

    return rot_stream


def _find_start_end_time(stream):
    """
    find the start and end time of a specific stream.

    :param stream:
    :return starttime, endtime:
    """

    starttime = []
    endtime = []

    for tr in stream:
        starttime.append(tr.stats.starttime)
        endtime.append(tr.stats.endtime)

    return max(starttime), min(endtime)


def remove_response(trace, resp_path, pre_filt=(0.01, 0.02, 8, 9), output="VEL", format="XML"):
    """
    Romove instrumental response

    :param trace:
    :param resp_path:
    :param pre_filt:
    :param output:
    :param format:
    :return:

    .. Note::
        currently only support XML DATALESS RESP format.
    """

    station_id = ".".join([trace.stats.network, trace.stats.station])
    trace_id = trace.id
    tr = trace.copy()

    if format.upper()in ["XML", "DATALESS", "RESP"]:
        if format.upper() == "RESP":
            resp_file = glob.glob(resp_path + "/*%s*" % trace_id)
        else:
            resp_file = glob.glob(resp_path + "/*%s*" % station_id)
        if len(resp_file) < 1:
            logging.error("Cannot find Response file: ", station_id)
        inv = read_inventory(resp_file[0])
        tr.remove_response(inventory=inv, pre_filt=pre_filt, output=output)
    elif method == "SACPZ":
        # from obspy.io.sac.sacpz import attach_paz
        # attach_paz(tr, paz_file="SACPZ.AU.WRKA.--.BHZ")
        # https://docs.obspy.org/packages/obspy.io.sac.sacpz.attach_paz.html#obspy.io.sac.sacpz.attach_paz
        logging.warn("The SACPZ format not supported currently.")
        pass
    else:
        logging.warn("other RESP format not supported: ", method)
        pass

    return tr
